#!/usr/bin/perl -w
###############################################################################
# This script is copyright(c) 2002 by asdfGroup.com, Inc. All rights reserved.
# It is subject to the license agreement that can be found at the following URL
#   http://www.asdfgroup.com/license.html
# Please contact our staff if you will have any questions:
#   mailto:staff@asdfgroup.com
###############################################################################

=head1 NAME

B<POWER::SQL> - Generate basic SQL queries

=head1 SYNOPSIS

 use POWER::SQL;

=head1 DESCRIPTION

=head2 

=over 4

=item *

       PRIMARY KEY.

=item *

   (JOIN)  ,   - 
    PRIMARY KEY     
.     .

=item *

   "_"   ,  
(         ).
    CGI  %Q    
,   "_".

=item *

    "-" .    
-order  -group .

=item *

   "__"     
   .

=item *

 Insert()  Update()    
_cdate (  )  _mdate (   )
  ,      .

=item *

    count(*),     
__count .

=back

=head1 METHODS

    POWER::SQL     DBI,  
      DBI: 

$row = $dbh->Row("SELECT * FROM Customers WHERE id_cust=?",555);

=over 4

=cut

package POWER::SQL;

use strict;
use vars qw($VERSION $dbh);

$VERSION = "1.03";

sub dbq { join ",", map { $dbh->quote($_) } ref $_[0] ? @{$_[0]} : $_[0] }
sub is_arr {ref $_[2] eq "ARRAY"}
sub is_int {$_[2] =~ m/^-?\d+ +(?:SECOND|MINUTE|HOUR|DAY|MONTH|YEAR)$/}
my %Func = (
    "eq"        => sub { ref $_[0] ? "$_[1] IN (".&dbq.")"     : "$_[1] =  ".&dbq },
    "ne"        => sub { ref $_[0] ? "$_[1] NOT IN (".&dbq.")" : "$_[1] != ".&dbq },
    "lt"        => sub { "$_[1] <  ".&dbq },
    "gt"        => sub { "$_[1] >  ".&dbq },
    "le"        => sub { "$_[1] <= ".&dbq },
    "ge"        => sub { "$_[1] >= ".&dbq },
    "date_eq"   => sub { &is_int and "$_[1] =  DATE_ADD(NOW(), INTERVAL $_[0])" },
    "date_ne"   => sub { &is_int and "$_[1] != DATE_ADD(NOW(), INTERVAL $_[0])" },
    "date_lt"   => sub { &is_int and "$_[1] <  DATE_ADD(NOW(), INTERVAL $_[0])" },
    "date_gt"   => sub { &is_int and "$_[1] >  DATE_ADD(NOW(), INTERVAL $_[0])" },
    "date_le"   => sub { &is_int and "$_[1] <= DATE_ADD(NOW(), INTERVAL $_[0])" },
    "date_ge"   => sub { &is_int and "$_[1] >= DATE_ADD(NOW(), INTERVAL $_[0])" },
    "like"      => sub { "$_[1] LIKE ".&dbq },
    "notlike"   => sub { "$_[1] NOT LIKE ".&dbq },
    "contain"   => sub { "$_[1] LIKE ".dbq('%'.$_[0].'%') },
    "notcontain"=> sub { "$_[1] NOT LIKE ".dbq('%'.$_[0].'%') },
    "inc"       => sub { "$_[1] = $_[1] + ".&dbq },
    );
my %DESC;

=item B<P2SQL>( $table, \%P )

=item B<P2SQL>( \@tables, \%P )

 $table  @tables   ,    ,  %P
  ,       $table  @tables.
   ( {ID}, {Field}, ...)    "."
    .

   :

=over 4

=item {DB}

      

=item {Desc}

       
  "" => $dbh->All("DESC ")

=item {Table}

   

=item {INNER}

   INNER JOIN

=item {LEFT}

   LEFT JOIN

=item {ID}

  PRIMARY KEY  {Table}

=item {FieldList}

  "=, 2=2"  
%P   ""

=item {FieldAND}

a-la {FieldList},     "AND" 
 -    "__"
        "1"

=item {Update}

a-la {FieldAND},      "" 
 primary key,   .

=item {Order}

  " ASC, 2 DESC"  ""

=item {Group}

a-la {Order}

=back

=cut

sub DBI::db::P2SQL {
    (local $dbh, my $tables, my $P) = @_;
    my %SQL;
    #        
    #  $SQL{DB}
    $DESC{$dbh} ||= {map {$_=>undef} @{$dbh->selectcol_arrayref("SHOW TABLES")}};
    $SQL{DB} = [ keys %{$DESC{$dbh}} ];
    #      ,     
    my @tables = grep { exists $DESC{$dbh}{$_} } ref $tables ? @$tables : $tables;
    # ,      
    return %SQL unless @tables;
    #      
    for my $t (@tables) {
        unless ($DESC{$dbh}{$t}) {
            (my $sth = $dbh->prepare("DESC $t"))->execute();
            $DESC{$dbh}{$t} = $sth->fetchall_arrayref({});
        }
    }
    #  $SQL{Table}
    $SQL{Table} = $tables[0];
    #  $SQL{Desc}   
    $SQL{Desc}{ $SQL{Table} } = $DESC{$dbh}{ $SQL{Table} };
    #  $SQL{ID}
    $SQL{ID} = $SQL{Desc}{ $SQL{Table} }[0]{Field};
    #  ,     
    my @join = $tables[0];
    my @field = map {{ map {$_->{Field}=>1} @{$DESC{$dbh}{$_}} }} @tables;
    for my $right (1..$#tables) {
        my $rkey = $DESC{$dbh}{ $tables[$right] }[0]{Field};
        for my $left (0..$right-1) {
            my $lkey = $DESC{$dbh}{ $tables[$left] }[0]{Field};
            my $key = $field[$left]{$rkey}  ? $rkey : 
                      $field[$right]{$lkey} ? $lkey : next;
            push @join, "JOIN $tables[$right]".
                " ON ($tables[$right].$key=$tables[$left].$key)";
            #  $SQL{Desc}    
            $SQL{Desc}{ $tables[$right] } = $DESC{$dbh}{ $tables[$right] };
            last;
        }
    }
    #  $SQL{INNER}  $SQL{LEFT}
    $SQL{INNER} = join " INNER ", @join;
    $SQL{LEFT}  = join " LEFT ",  @join;
    #       
    my %qualify;
    for my $t (@tables) {
        next unless exists $SQL{Desc}{$t}; #     join
        $qualify{$_} ||= "$t.$_" for map {$_->{Field}} @{$SQL{Desc}{$t}};
    }
    #  $SQL{FieldList}, $SQL{FieldAND}, $SQL{Update}
    for my $f (keys %$P) {
        my ($name, $func) = $f =~ /^(.*?)(?:__(.*))?$/;
        next unless my $qname = $qualify{$name};            #   
        next unless my $sub   = $Func{$func || "eq"};       #   
        my $is_list = ref($P->{$f}) && ($func ne "eq") && ($func ne "ne");
      for my $value ($is_list ? @{$P->{$f}} : $P->{$f}) {
        next unless my $expr  = $sub->($value, $qname);     # . .
        $SQL{FieldList} .= ", $expr" if (not $func or $func eq "inc");
        $SQL{FieldAND}  .= " AND $expr";
        $SQL{Update}    .= " AND $expr" if (($func and $func ne "inc") or
            $name eq $SQL{ID});
      }
    }
    $SQL{FieldList} =~ s/^, //;
    $SQL{FieldAND}  =~ s/^ AND //; $SQL{FieldAND} ||= "1";
    $SQL{Update}    =~ s/^ AND //; $SQL{Update}   ||= "1";
    #  $SQL{Order}
    for my $order (ref $P->{-order} ? @{$P->{-order}} : $P->{-order}) {
        next unless $order =~ /^(\w+)(?: ASC| DESC)?$/;
        next unless $qualify{$1};
        $SQL{Order} = join ", ", ($SQL{Order}||()), $order;
    }
    #  $SQL{Group}
    for my $group (ref $P->{-group} ? @{$P->{-group}} : $P->{-group}) {
        next unless $group =~ /^(\w+)(?: ASC| DESC)?$/;
        next unless $qualify{$1};
        $SQL{Group} = join ", ", ($SQL{Group}||()), $group;
    }
    return %SQL;
}

=item B<Insert>( $table, \%P )

INSERT INTO ?table SET ?set

     "_cdate",  "_mdate"     NOW().
 LAST_INSERT_ID()     undef().

=cut

sub DBI::db::Insert {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, {"_cdate"=>undef, "_mdate"=>undef, %$P});
    return unless $SQL{Table} and $SQL{FieldList};
    $SQL{FieldList} =~ s/(_[cm]date\s*)=\s*NULL/$1=NOW()/g;
    return $dbh->do("INSERT INTO $SQL{Table} SET $SQL{FieldList}")
        ? $dbh->selectrow_array("SELECT LAST_INSERT_ID()") : undef;
}

=item B<Update>( $table, \%P )

UPDATE ?table SET ?set WHERE ?id=... AND ?where

     "_mdate",  '_mdate=NOW()'.
   SET     "",    WHERE  
  "__"  primary_key     "".
 -    undef()  .

=cut

sub DBI::db::Update {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, {"_mdate"=>undef, %$P});
    return unless $SQL{Table} and $SQL{FieldList};
    $SQL{FieldList} =~ s/(_mdate\s*)=\s*NULL/$1=NOW()/g;
    return $dbh->do("UPDATE $SQL{Table} SET $SQL{FieldList} WHERE $SQL{Update}");
}

=item B<Replace>( $table, \%P )

REPLACE ?table SET ?set

     "_cdate"  "_mdate",     NOW().
   SET     "".
    REPLACE  undef().

=cut

sub DBI::db::Replace {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, {"_cdate"=>undef, "_mdate"=>undef, %$P});
    return unless $SQL{Table} and $SQL{FieldList};
    $SQL{FieldList} =~ s/(_[cm]date\s*)=\s*NULL/$1=NOW()/g;
    return $dbh->do("REPLACE INTO $SQL{Table} SET $SQL{FieldList}");
}

=item B<Delete>( $tables, \%P )

DELETE FROM ?table WHERE ?where

    .  $tables  undef(),  
         %P .

=cut

sub DBI::db::Delete {
    my ($dbh, $table, $P) = @_;
    if (not defined $table) {
        return unless my @field = map {s/__.*//;$_} keys %$P;
        my %SQL = $dbh->P2SQL({$dbh->P2SQL()}->{DB});
        for my $t (@{$SQL{DB}}) {
            my %seen;
            push @$table, $t if @field == grep {++$seen{$_}==2} 
                @field, map {$_->{Field}} @{$SQL{Desc}{$t}};
        }
    }
    for my $t (ref $table ? @$table : $table) {
        my %SQL = $dbh->P2SQL($t, $P);
        next unless $SQL{Table};
        $dbh->do("DELETE FROM $SQL{Table} WHERE $SQL{FieldAND}");
    }
}

=item B<ID>( $tables, \%P )

SELECT primary_key FROM ?inner WHERE ?where [ORDER BY ?order]

   Col() .

=cut

sub DBI::db::ID {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, $P);
    return unless $SQL{Table};
    $dbh->Col("SELECT $SQL{ID} FROM $SQL{INNER} WHERE $SQL{FieldAND}".
        ($SQL{Order} ? " ORDER BY $SQL{Order}" : ""));
}

=item B<Count>( $tables, \%P )

SELECT count(*) FROM ?inner WHERE ?where

   Col() .

=cut

sub DBI::db::Count {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, $P);
    return unless $SQL{Table};
    $dbh->Col("SELECT count(*) __count FROM $SQL{INNER} WHERE $SQL{FieldAND}");
}

=item B<Select>( $tables, \%P )

SELECT * FROM ?inner WHERE ?where [ORDER BY ?order]
SELECT *, ?group, count(*) __count FROM ?inner WHERE ?where GROUP BY ?group
    [ORDER BY ?order]

   All()      Row()  .

=cut

sub DBI::db::Select {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, $P);
    return unless $SQL{Table};
    my $SQL = "SELECT *".
        ($SQL{Group} ? ", $SQL{Group}, count(*) __count" : "").
        " FROM $SQL{INNER} WHERE $SQL{FieldAND}".
        ($SQL{Group} ? " GROUP BY $SQL{Group}" : "").
        ($SQL{Order} ? " ORDER BY $SQL{Order}" : "");
    return wantarray ? $dbh->All($SQL) : $dbh->Row($SQL);
}

=item B<All>($sql[, @bind])

Execute $sql and return array of hashrefs to row data.
Example:

 [ { id => 1, user => "root" },
   { id => 2, user => "daemon" },
   { id => 3, user => "system" },
 ]

=cut

sub DBI::db::All {
    my ($dbh, $sql, @bind) = @_;
    (my $sth = $dbh->prepare($sql))->execute(@bind);
    @{ $sth->fetchall_arrayref({}) };
}

=item B<Row>($sql[, @bind])

Execute $sql and return reference to hash with row data or undef.
Example:

 { id => 1, user => "root" }

=cut

sub DBI::db::Row {
    my ($dbh, $sql, @bind) = @_;
    (my $sth = $dbh->prepare($sql))->execute(@bind);
    $sth->fetchrow_hashref();
}

=item B<Col>($sql[, @bind])

Execute $sql and return array of column data in list context or
first cell in scalar context.

=cut

sub DBI::db::Col {
    my ($dbh, $sql, @bind) = @_;
    my @res = @{ $dbh->selectcol_arrayref($sql, undef, @bind) };
    return wantarray ? @res : $res[0];
}

=back

=head1 FUNCTIONS for FIELDS

,         "__":

=over 8

=item B<eq>, B<ne>, B<lt>, B<gt>, B<le>, B<ge>

name = != < > <= >= value

 value        B<eq>  B<ne>:

name [NOT] IN (value0,value1,...)

value must be /^\w+$/

=item B<date_eq>, B<date_ne>, B<date_lt>, B<date_gt>, B<date_le>, B<date_ge>

name = != < > <= >= DATE_ADD(NOW(), INTERVAL value)

value must be /^-?\d+ (?:SECOND|MINUTE|HOUR|DAY|MONTH|YEAR)$/

=item B<like>, B<notlike>

name [NOT] LIKE value

=item B<contain>, B<notcontain>

name [NOT] LIKE '%'.value.'%'

=item B<inc>

name = name + value

  Update()    SET,   WHERE.
     Insert()  Replace() .

=back

=head1 HISTORY

 1.03         22 16:48:19 EET 2002
        -   __B<inc>

 1.02         20 13:50:16 EET 2002
        -   __B<btw>
        -  IN   __B<eq>  __B<ne>
        -         __B<eq> 
           __B<ne>      
            

 1.01         17 23:19:20 EET 2002
        - bugfix:   -order  -group  ASC|DESC
        - bugfix:   -order  -group 
            
        -   __B<like>  __B<notlike>
        -   __B<contain>  __B<notcontain>

 1.00         15 16:52:30 EET 2002
        - first release

=head1 AUTHORS

Alex Efros <powerman@asdfGroup.com>
Nikita Savin <nikita@asdfGroup.com>

=cut

1;
